/** 
 * \file adg_dcel_mesh.h
 *  
 * \brief  Definition and implementation  of the  class Adg_dcel_mesh,
 * which implements  the abstract interface Generic_mesh  using a DCEL
 * data structure.
 *
 * \author
 * Mario Augusto de Souza Lizier \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * lizier at gmail (dot) com
 *
 * \author
 * Marcelo Ferreira Siqueira
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * mfsiqueira at gmail (dot) com
 *
 * \version 1.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef ADG_DCEL_MESH_H
#define	ADG_DCEL_MESH_H

#include "surface.h"                // dcel::Surface 
#include "generic_mesh.h"           // agd::Generic_mesh
#include "adg_vertex_attributes.h"  // Adg_vertex_attributes

#include <cassert>                  // assert


/** 
* \defgroup ADGFROMDCELNameSpace Namespace adgfromdcel.
* @{
*/

/**
 * \namespace adgfromdcel
 *
 * \brief The namespace adgfromdcel contains the definition of classes
 * Adg_dcel_mesh and Adg_vertex_attributes,  which allow us to compute
 * discrete geodesics  defined over triangle surface  meshes stored in
 * the DCEL data structure.
 */

namespace adgfromdcel {

  using dcel::Surface ;
  using adg::Generic_mesh ;
  using adg::Curve_point ;

  /**
   * \typedef mySurface
   *
   * \brief Definition of a type name for the surface mesh.
   */
  typedef dcel::Surface< Adg_vertex_attributes > mySurface ;

  /**
   * \class Adg_dcel_mesh
   *
   * \brief This class  implements the abstract interface Generic_mesh
   * using  the DCEL  data structure,  and  thus allows  us to  define
   * discrete  geodesics over  triangle surface  meshes stored  in the
   * DCEL.
   */
  class Adg_dcel_mesh : public Generic_mesh< mySurface > {
  public:

    /**
     * \fn Adg_dcel_mesh( mySurface* m )
     *
     * \brief Creates an instance of this class.
     *
     * \param m Pointer to a mesh stored in a DCEL.
     */    
    Adg_dcel_mesh( mySurface* m ) : _mesh( m )
    {
      assert( _mesh != 0 ) ;
    }


    /**
     * \fn ~Adg_dcel_mesh()
     *
     * \brief Destroys an instance of this class.
     */
    virtual ~Adg_dcel_mesh()
    {}


    /**
     * \fn inline Vertex* get_org( Halfedge* h ) const
     *
     * \brief Returns  the origin vertex  of a given half-edge  of the
     * mesh.
     *
     * \param h Pointer to a half-edge of the mesh.
     *
     * \return A pointer to the origin vertex of a given half-edge.
     */
    inline Vertex* get_org( Halfedge* h )  const 
    {
      return h->get_origin() ;
    }


    /**
     * \fn inline Edge* get_edge( Halfedge* h ) const
     *
     * \brief Returns the  edge a given half-edge of  the mesh belongs
     * to.
     *
     * \param h Pointer to a half-edge of the mesh.
     *
     * \return A pointer to the edge a given half-edge belongs to.
     */
    inline Edge* get_edge( Halfedge* h ) const 
    {
      return h->get_edge() ;
    }


    /**
     * \fn inline Face* get_face( Halfedge* h ) const
     *
     * \brief Returns the  face a given half-edge of  the mesh belongs
     * to.
     *
     * \param h Pointer to a half-edge of the mesh.
     *
     * \return A pointer to the face a given half-edge belongs to.
     */
    inline Face* get_face( Halfedge* h ) const
    {
      return h->get_face() ;
    }


    /**
     * \fn inline Halfedge* get_prev( Halfedge* h ) const
     *
     * \brief Returns the half-edge that precedes a given half-edge of
     * the mesh  in the  face cycle of  half-edges that  contains both
     * half-edges.
     *
     * \param h Pointer to a half-edge of the mesh.
     *
     * \return  A pointer  to  the half-edge  that  precedes a  given
     * half-edge in their common face half-edge cycle.
     *
     */
    inline Halfedge* get_prev( Halfedge* h ) const 
    {
      return h->get_prev() ;
    }


    /**
     * \fn inline Halfedge* get_next( Halfedge* h ) const
     *
     * \brief Returns the half-edge that succeeds a given half-edge of
     * the mesh  in the  face cycle of  half-edges that  contains both
     * half-edges.
     *
     * \param h Pointer to a half-edge of the mesh.
     *
     * \return  A pointer  to  the half-edge  that  succeeds a  given
     * half-edge in their common face half-edge cycle.
     */
    inline Halfedge* get_next( Halfedge* h ) const 
    {
      return h->get_next() ;
    }


    /**
     * \fn Halfedge* get_mate( Halfedge* h ) const
     *
     * \brief  Returns the  mate  of  a given  half-edge.
     *
     * \param h Pointer to a half-edge of the mesh.
     *
     * \return A pointer to the mate half-edge of a given half-edge.
     *
     */
    inline Halfedge* get_mate( Halfedge* h ) const 
    { 
      return h->get_mate() ;
    }


    /**
     * \fn inline Halfedge* get_halfedge( Face* face ) const
     *
     * \brief Returns  the first half-edge of the  cycle of half-edges
     * of a given face of the mesh.
     *
     * \param face Pointer to a face of the mesh.
     *
     * \return A pointer to the first half-edge of a given face.
     */
    inline Halfedge* get_halfedge( Face* face ) const
    {
      return face->get_halfedge() ;
    }


    /**
     * \fn inline Halfedge* get_halfedge( Vertex* vertex ) const
     *
     * \brief Returns one half-edge with origin at a given vertex.  It
     * is  assumed  that  this  method  will always  return  the  same
     * half-edge  (as  many  half-edges  may  share  the  same  origin
     * vertex).
     *
     * \param vertex Pointer to a vertex of the mesh.
     *
     * \return  A pointer  to one  half-edge  with origin  at a  given
     * vertex.
     */
    inline Halfedge* get_halfedge( Vertex* vertex ) const
    {
      return vertex->get_halfedge() ;
    }


    /**
     * \fn inline VertexIterator vertices_begin() const
     *
     * \brief Returns a vertex iterator set to the initial vertex of a
     * vertex sequence of the mesh.
     *
     * \return A vertex iterator set to the initial vertex of a vertex
     * sequence of the mesh.
     */
    inline VertexIterator vertices_begin() const 
    {
      return _mesh->vertices_begin() ;
    }


    /**
     * \fn inline bool is_done( const VertexIterator& iterator ) const
     *
     * \brief Returns  a logic value  true if a given  vertex iterator
     * has  reached  the  end  of  a  vertex  sequence  of  the  mesh;
     * otherwise, it returns the logic value false.
     *
     * \param iterator A vertex iterator.
     *
     * \return  A logic  value true  if  a given  vertex iterator  has
     * reached the end of a vertex sequence of the mesh; otherwise, it
     * returns the logic value false.
     */
    inline bool is_done( const VertexIterator& iterator ) const
    {
      return ( iterator == _mesh->vertices_end() ) ;
    }


    /**
     * \fn inline void move_forward( VertexIterator& iterator ) const
     *
     * \brief Makes  the iterator point  to the vertex  succeeding its
     * current vertex in a vertex sequence of the mesh.
     *
     * \param iterator A reference to a vertex iterator.
     */
    inline void move_forward( VertexIterator& iterator ) const
    {
      ++iterator ;
    }


    /**
     * \fn inline Vertex* get_vertex( const VertexIterator& iterator ) const
     *
     * \brief Returns  the current vertex  of a given  vertex iterator
     * for a vertex sequence of the mesh.
     *
     * \param iterator A reference to a vertex iterator.
     *
     * \return  A pointer  to the  current  vertex of  a given  vertex
     * iterator for a vertex sequence of the mesh.
     */
    inline Vertex* get_vertex( const VertexIterator& iterator ) const
    {
      return *iterator ;
    }


    /**
     * \fn virtual EdgeIterator edges_begin() const
     *
     * \brief Returns an  edge iterator set to the  initial edge of an
     * edge sequence of the mesh.
     *
     * \return An  edge iterator  set to the  initial edge of  an edge
     * sequence of the mesh.
     *
     */
    inline EdgeIterator edges_begin() const 
    {
      return _mesh->edges_begin() ;
    }


    /**
     * \fn inline bool is_done( const EdgeIterator& iterator ) const
     *
     * \brief Returns A logic value  true if a given edge iterator has
     * reached the end of an  edge sequence of the mesh; otherwise, it
     * returns the logic value false.
     *
     * \param iterator An edge iterator.
     *
     * \return A logic value true if a given edge iterator has reached
     * the end of an edge  sequence of the mesh; otherwise, it returns
     * the logic value false.
     */
    inline bool is_done( const EdgeIterator& iterator ) const
    {
      return ( iterator == _mesh->edges_end() ) ;
    }


    /**
     * \fn inline void move_forward( EdgeIterator& iterator ) const
     *
     * \brief  Makes the  iterator point  to the  edge  succeeding its
     * current edge in an edge sequence of the mesh.
     *
     * \param iterator A reference to an edge iterator.
     */
    inline void move_forward( EdgeIterator& iterator ) const 
    {
      ++iterator ;
    }


    /**
     * \fn inline Edge* get_edge( const EdgeIterator& iterator ) const
     *
     * \brief Returns the current edge of a given edge iterator for an
     * edge sequence of the mesh.
     *
     * \param iterator A reference to an edge iterator.
     *
     * \return A pointer to the  current edge of a given edge iterator
     * for an edge sequence of the mesh.
     */
    inline Edge* get_edge( const EdgeIterator& iterator ) const
    {  
      return *iterator ;
    }


    /**
     * \fn inline FaceIterator faces_begin() const
     *
     * \brief Returns  a face  iterator set to  the initial face  of a
     * face sequence of the mesh.
     *
     * \return  A face  iterator set  to the  initial face  of  a face
     * sequence of the mesh.
     */
    inline FaceIterator faces_begin() const 
    {
      return _mesh->faces_begin() ;
    }


    /**
     * \fn virtual bool is_done( const FaceIterator& iterator ) const
     *
     * \brief Returns A logic value  true if a given face iterator has
     * reached the end  of a face sequence of  the mesh; otherwise, it
     * returns the logic value false.
     *
     * \param iterator A face iterator.
     *
     * \return A logic value true if a given face iterator has reached
     * the end of  a face sequence of the  mesh; otherwise, it returns
     * the logic value false.
     */
    inline bool is_done( const FaceIterator& iterator ) const 
    {
      return ( iterator == _mesh->faces_end() ) ;
    }


    /**
     * \fn inline void move_forward( FaceIterator& iterator ) const
     *
     * \brief  Makes the  iterator point  to the  face  succeeding its
     * current face in a face sequence of the mesh.
     *
     * \param iterator A reference to a face iterator.
     */
    inline void move_forward( FaceIterator& iterator ) const 
    {
      ++iterator ;
    }


    /**
     * \fn inline Face* get_face( const FaceIterator& iterator ) const
     *
     * \brief Returns the current face  of a given face iterator for a
     * face sequence of the mesh.
     *
     * \param iterator A reference to a face iterator.
     *
     * \return A pointer to the  current face of a given face iterator
     * for a face sequence of the mesh.
     */
    inline Face* get_face( const FaceIterator& iterator ) const 
    {
      return *iterator ;
    }


    /**
     * \fn inline void get_halfedges( Edge* e, Halfedge*& h1, Halfedge*& h2 ) const
     *
     * \brief Returns both halfedges that define a given edge.
     *
     * \param e A reference to an edge iterator.
     * \param h1 A pointer to receive the first halfedge.
     * \param h2 A pointer to receive the second halfedge.
     *
     * \return Two pointers to the halfedges of a given edge.
     */
    inline void get_halfedges( 
			      Edge* e , 
			      Halfedge*& h1 ,
			      Halfedge*& h2 
			     ) 
      const 
    {
      h1 = e->get_first_halfedge() ;
      h2 = e->get_second_halfedge() ;
    }


    /**
     * \fn inline void get_coords( Vertex* v, double& x , double& y , double& z ) const
     *
     * \brief Returns the three coordinates of a given vertex.
     *
     * \param v A reference to the given vertex.
     * \param x Returns X coordinate.
     * \param y Returns Y coordinate.
     * \param z Returns Z coordinate.
     */
    inline void get_coords( 
			   Vertex* v ,
			   double& x ,
			   double& y ,
			   double& z 
			  ) 
      const 
    {
      x = v->x() ;
      y = v->y() ;
      z = v->z() ;
    }


    /**
     * \fn inline void set_coords( Vertex* v, double x, double y, double z )
     *
     * \brief Define the three coordinates of a given vertex.
     *
     * \param v A reference to the given vertex;
     * \param x X coordinate;
     * \param y Y coordinate;
     * \param z Z coordinate;
     *
     */
    inline void set_coords( 
			   Vertex* v ,
			   double x ,
			   double y ,
			   double z 
			  ) 
    {
      v->set_x_coord( x ) ;
      v->set_y_coord( y ) ;
      v->set_z_coord( z ) ;
    }


    /**
     * \fn inline bool is_alive( Vertex* v ) const
     *
     * \brief Checks if the vertex is in the alive set.
     *
     * \param v A pointer to a given vertex.
     *
     * \return  The logic value  true if  the given  vertex is  in the
     * alive set, and the logic value false otherwise.
     */
    inline bool is_alive( Vertex* v ) const 
    {
      return v->get_attributes().alive() ;
    }


   /**
    * \fn inline void set_alive( Vertex* v, bool a )
    *
    * \brief  Assigns  a  given   logic  value  to  the  vertex  flag
    * (supposedly a vertex attribute) that indicates if the vertex is
    * in the alive set.
    *
    * \param v A pointer to a given vertex.
    * \param a A logic value (i.e., true or false).
    */
    inline void set_alive( Vertex* v , bool a ) {
      v->get_attributes().set_alive( a ) ;
    }


    /**
     * \fn inline bool is_close( Vertex* v ) const
     *
     * \brief Checks if the vertex is in the alive close.
     *
     * \param v A pointer to a given vertex.
     *
     * \return  The logic value  true is  the given  vertex is  in the
     * close set, and the logic value false otherwise.
     */
    bool is_close( Vertex* v ) const 
    {
      return v->get_attributes().close() ;
    }


    /**
     * \fn inline void set_close( Vertex* v, bool a )
     *
     * \brief  Assigns  a  given   logic  value  to  the  vertex  flag
     * (supposedly a vertex attribute) that indicates if the vertex is
     * in the close set.
     *
     * \param v A pointer to a given vertex.
     * \param a A logic value (i.e., true or false).
     */
    inline void set_close( 
			  Vertex* v ,
			  bool a 
			 ) 
    {
      v->get_attributes().set_close( a ) ;
    }


    /**
     * \fn inline double get_dist( Vertex* v ) const
     *
     * \brief Returns  the distance  function value associated  with a
     * given vertex.
     *
     * \param v A pointer to a given vertex.
     *
     * \return The distance funtion value at the given vertex.
     */
    inline double get_dist( Vertex* v ) const 
    {
      return v->get_attributes().dist() ;
    }
    

    /**
     * \fn inline void set_dist( Vertex* v, double d )
     *
     * \brief Assigns a given  distance value to the distance function
     * attribute of a given vertex.
     *
     * \param v A pointer to a given vertex.
     * \param d A distance value (i.e., a real number).
     */
    inline void set_dist( 
			 Vertex* v ,
			 double d 
			 ) 
    {
      v->get_attributes().set_dist( d ) ;
    }


    // ---------------------------------------------------------------
    //
    // Private data members.
    //
    // ---------------------------------------------------------------

    mySurface* _mesh ; ///< Pointer to the actual mesh.

  } ;

}

/** @} */ //end of group class.

#endif	/* ADG_DCEL_MESH_H */

