/** 
 * \file adg_vertex_attributes.h
 *  
 * \brief  Definition   of  the  class   Adg_vertex_attributes,  which
 * represents  the  set   of  discrete  geodesic  computation  related
 * attributes associated  with the vertices of the  DCEL triangle mesh
 * data structure.
 *
 * \author
 * Mario Augusto de Souza Lizier \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * lizier at gmail (dot) com
 *
 * \version 1.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef ADG_ATTRIBUTES_H
#define	ADG_ATTRIBUTES_H

#include <cmath>  // HUGE_VAL


/** 
* \defgroup ADGFROMDCELNameSpace Namespace adgfromdcel.
* @{
*/

/**
 * \namespace adgfromdcel
 *
 * \brief The namespace adgfromdcel contains the definition of classes
 * Adg_dcel_mesh and Adg_vertex_attributes,  which allow us to compute
 * discrete geodesics  defined over triangle surface  meshes stored in
 * the DCEL data structure.
 */

namespace adgfromdcel {

  /**
   * \class Adg_vertex_attributes
   *
   * \brief  This  class  represents  the  set  of  discrete  geodesic
   * computation  related attributes associated  with the  vertices of
   * the DCEL triangle mesh data  structure. All atributes are used by
   * the implementation  of the Fast  Marching Method (FMM),  which in
   * turn is used by the algorithm that computes an approximation to a
   * geodesic.
   *
   * \sa Adg_dcel_mesh, Geodesics
   */
  class Adg_vertex_attributes {
  public:
    // ---------------------------------------------------------------
    //
    // Public methods.
    //
    // ---------------------------------------------------------------

    /**
     * \fn Adg_vertex_attributes( bool a , double d , bool c )
     *
     * \brief Creates an  instance of this class.
     *
     * \param a A  flag to indicate whether the  vertex that owns this
     * attribute  is in  the alive  set defined  by the  Fast Marching
     * Method (FMM).
     * \param d The  value computed by the Fast  Marching Method (FMM)
     * for  the  distance  function  at  the  vertex  that  owns  this
     * attribute.
     * \param c A  flag to indicate whether the  vertex that owns this
     * attribute  is in  the close  set defined  by the  Fast Marching
     * Method (FMM).
     */    
    Adg_vertex_attributes(
			  bool a = false ,
			  double d = -HUGE_VAL ,
			  bool c = false ) 
    {

        _a = a;
        _d = d;
        _c = c;
    }


    /**
     * \fn inline bool alive() const
     *
     * \brief Returns  the value of  the alive set membership  flag of
     * the vertex associated with this attribute.
     *
     * \return  The logic  value true  if  the vertex  that owns  this
     * attribute  is in  the  alive  set, and  the  logic value  false
     * otherwise.
     */
    inline bool alive() const {
      return _a ;
    }


    /**
     * \fn double dist() const
     *
     * \brief Returns the distance  function value associated with the
     * vertex that owns this attribute.
     *
     * \return  The distance  funtion value  at the  vertex associated
     * with this attribute.
     */
    inline double dist() const {
      return _d ; 
    }


    /**
     * \fn inline bool close() const
     *
     * \brief Returns  the value of  the close set membership  flag of
     * the vertex associated with this attribute.
     *
     * \return The logic value true is the vertex associated with this
     * attribute  is in  the  close  set, and  the  logic value  false
     * otherwise.
     */
    inline bool close() const 
    { 
      return _c ;
    }


    /**
     * \fn inline void set_alive( bool a )
     *
     * \brief Assigns a given logic  value to the alive set membership
     * flag of the vertex associated with this attribute.
     *
     * \param a A logic value (i.e., true or false).
     */
    void set_alive( bool a ) 
    { 
      _a = a ; 
    }


    /**
     * \fn void set_dist( double d )
     *
     * \brief  Assigns a given  value to  the distance  function value
     * associated with the vertex that owns this attribute.
     *
     * \param d A distance value (i.e., a real number).
     */
    void set_dist( double d ) 
    { 
      _d = d ; 
    }

    /**
     * \fn void set_close( bool c )
     *
     *
     * \brief Assigns a given logic  value to the close set membership
     * flag of the vertex associated with this attribute.
     *
     * \param c A logic value (i.e., true or false).
     */
    void set_close( bool c ) 
    { 
      _c = c ; 
    }

    
private:

    // ---------------------------------------------------------------
    //
    // Private data members.
    //
    // ---------------------------------------------------------------

    bool _a ;  ///< The alive set membership flag.

    double _d ;  ///< The distance function value.

    bool _c ;  ///< The close set membership flag.

  } ;


}

/** @} */ //end of group class.

#endif	/* ADG_ATTRIBUTES_H */
