/** 
 * \file lexer.cpp
 *  
 * \brief  Implementation  of  the  class Lexer,  which  represents  a
 * lexical analyzer for scanning tokens  from an OFF file describing a
 * triangle mesh.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date July 2009
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "lexer.h"

#include <sstream>
#include <cctype>
#include <cstdlib>
#include <cassert>


/** 
* \defgroup OFFLIBNameSpace Namespace offlib.
* @{
*/

/**
* \namespace offlib
*
* \brief   The   namespace   offlib   contains  the   definition   and
* implementation of  classes for  reading and writing  the topological
* and geometric information of triangle meshes from / to OFF files.
*/

namespace offlib {

  /**
   * \fn bool Lexer::get_string( std::string& s )
   *
   * \brief Reads in a string from the input stream.
   *
   * \param s A string read in from the input stream.
   */
  bool
  Lexer::get_string( std::string& s ) 
  {
    if ( !skip_space() ) {
      return false;
    }

    char c;
    if ( !_realstr.get( c ) ) {
      return false ;
    }

    if ( isalpha( c ) ) {
      char str[ 256 ] ;
      int i = 0 ;
      do {
        str[i] = c ;

        if ( !_realstr.get( c ) ) {
          return false ;
        }

        increment_char() ;

        ++i ;
      } 
      while ( isalnum( c ) && ( i < 255 ) ) ;

      str[ i ] = '\0' ;

      assert( !isalnum( c ) ) ;

      s = str ;
      _realstr.putback( c ) ;
    }
    else {
      _realstr.putback( c ) ;

      return false ;
    }

    return true ;
  }


  /**
   * \fn bool Lexer::get_integer( int& x )
   *
   * \brief Reads in a non-negative integer from the input stream.
   *
   * \param x A reference to an integer.
   */
  bool
  Lexer::get_integer( int& x )
  {
    if ( !skip_space() ) {
      return false ;
    }

    char c ;
    if ( !_realstr.get( c ) ) {
      return false ;
    }

    if ( isdigit( c ) ) {
      char str[ 256 ] ;
      int i = 0 ;
      do {
        str[ i ] = c ;

        if ( !_realstr.get( c ) ) {
          return false ;
        }

        increment_char() ;

        ++i ;
      } 
      while ( isdigit( c ) && ( i < 255 ) ) ;

      str[ i ] = '\0' ;

      assert( !isdigit( c ) ) ;

      x = atoi( str ) ;

      _realstr.putback( c ) ;
    }
    else {
      _realstr.putback( c ) ;

      return false;
    }

    return true;
  }


  /**
   * \fn bool Lexer::get_double( double& d )
   *
   * \brief Reads in a double precision number from the input stream.
   *
   * \param d A reference to a double precision number.
   */
  bool
  Lexer::get_double( double& d )
  {
    if ( !skip_space() ) {
      return false ;
    }

    char c ;
    if ( !_realstr.get( c ) ) {
      return false ;
    }
   
    _realstr.putback( c ) ;

    assert( ( _realstr >> d ) ) ;

    return true;
  }


  /**
   * \fn bool Lexer::skip_space()
   *
   * \brief Reads in the input stream until a character or digit shows
   * up.
   */
  bool
  Lexer::skip_space() 
  {
    char c ;

    for ( ; ; ) {
      if ( !_realstr.get( c ) ) {
        return false ;
      }

      increment_char() ;

      if ( c == '\n' ) {
        increment_line() ;
      }

      if ( !isspace( c ) ) {
        _realstr.putback( c ) ;
        decrement_char() ;
        return true ;
      }
    }

    return true ;
  }

}

/** @} */ //end of group class.
