/** 
 * \file lexer.h
 *  
 * \brief Definition  of the class  Lexer, which represents  a lexical
 * analyzer for scanning tokens from an OFF file describing a triangle
 * mesh.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date July 2009
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef LEXER_H
#define LEXER_H

#include <iostream>
#include <string>


/** 
* \defgroup OFFLIBNameSpace Namespace offlib.
* @{
*/

/**
* \namespace offlib
*
* \brief   The   namespace   offlib   contains  the   definition   and
* implementation of  classes for  reading and writing  the topological
* and geometric information of triangle meshes from / to OFF files.
*/

namespace offlib {


  /**
   * \class Lexer
   *
   * \brief  This class  represents  a lexical  analyzer for  scanning
   * tokens from an OFF file describing a triangle mesh.
   */
  class Lexer {
  public:
    // ---------------------------------------------------------------
    //
    // Public methods.
    //
    // ---------------------------------------------------------------

    /**
     * \fn Lexer( std::istream& is )
     *
     * \brief Creates an instance of this class.
     *
     * \param is An input stream for an OFF file.
     */
    Lexer(std::istream& is ) : _realstr( is )
    {
      _line_counter = 0 ;
      _char_counter = 0 ;
    }


    /**
     * \fn bool get_string( std::string& s )
     *
     * \brief Reads in a string from the input stream.
     *
     * \param s A reference to a string.
     */
    bool get_string( std::string& s ) ;


    /**
     * \fn bool get_integer( int& x )
     *
     * \brief Reads in a non-negative integer from the input stream.
     *
     * \param x A reference to an integer.
     */
    bool get_integer( int& x ) ;


    /**
     * \fn bool get_double( double& d )
     *
     * \brief  Reads  in a  double  precision  number  from the  input
     * stream.
     *
     * \param d A reference to a double precision number.
     */
    bool get_double( double& d ) ;


    /**
     * \fn inline int get_line_counter() const
     *
     * \brief Returns the value of the file line counter.
     *
     * \return The value of the file line counter.
     */
    inline int get_line_counter() const
    { 
      return _line_counter ; 
    }


    /**
     * \fn inline int get_char_counter() const
     *
     * \brief Returns the value of the file char counter.
     *
     * \return The value of the file char counter.
     */
    int get_char_counter() const
    { 
      return _char_counter ; 
    }

  private:
    // ---------------------------------------------------------------
    //
    // Private methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn inline void increment_line()
     *
     * \brief Increments the file line counter. 
     */
    inline void increment_line()
    {
      ++_line_counter ;
      reset_char_counter() ;
    }


    /**
     * \fn inline void increment_char()
     *
     * \brief Increments the file char counter. 
     */
    inline void increment_char()
    {
      ++_char_counter ;
    }


    /**
     * \fn inline void decrement_char()
     *
     * \brief Decrements the file char counter. 
     */
    inline void decrement_char()
    {
      if ( _char_counter > 0 ) --_char_counter ;
    }


    /**
     * \fn inline void reset_char_counter()
     *
     * \brief Assigns the value zero to the file char counter.
     */
    inline void reset_char_counter()
    { 
      _char_counter = 0 ; 
    }


    /**
     * \fn bool skip_space()
     *
     * \brief Reads  in the  input stream until  a character  or digit
     * shows up.
     */
    bool skip_space() ;


    // ---------------------------------------------------------------
    //
    // Private data members
    //
    // ---------------------------------------------------------------

    std::istream& _realstr ;  ///< The input stream. 

    int _line_counter ;       ///< The file line counter.

    int _char_counter ;       ///< The file char counter.

  } ;

}

/** @} */ //end of group class.

#endif    // LEXER_H
