/** 
 * \file reader.cpp
 *  
 * \brief Implementation of the  class Reader, which represents a file
 * reader for reading  in a triangle surface mesh  information from an
 * OFF file.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date July 2009
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "reader.h"

#include <iostream>
#include <sstream>
#include <cassert>


/** 
* \defgroup OFFLIBNameSpace Namespace offlib.
* @{
*/

/**
* \namespace offlib
*
* \brief   The   namespace   offlib   contains  the   definition   and
* implementation of  classes for  reading and writing  the topological
* and geometric information of triangle meshes from / to OFF files.
*/

namespace offlib {

  using std::cout;
  using std::endl;

  /**
   * \fn Reader::read( unsigned& nv , double*& vset , unsigned& nf , unsigned*& fset ) 
   *
   * \brief Reads in the  topological and geometric information of the
   * surface mesh described in the input file.
   *
   * \param nv A  reference to the number of  vertices of the surface.
   * \param vset A reference to  an array with the vertex coordinates.
   * \param  nf A reference  to the  number of  faces of  the surface.
   * \param  fset  A  reference  to  an array  with  the  face  vertex
   * identifiers.
   */
  void
  Reader::read(
	       unsigned& nv ,
	       double*& vset ,
	       unsigned& nf ,
	       unsigned*& fset
	      )
  {
    std::filebuf fb ;

    assert( fb.open( _fname.c_str() , std::ios::in ) != 0 ) ;

    std::istream istr( &fb ) ;

    _is = new Lexer( istr ) ;

    vset = 0 ;
    fset = 0 ;

    read_header( nv , nf ) ;

    read_vertices( nv , vset ) ;

    read_faces( nf , nv , fset ) ;

    fb.close();
  }


  /**
   * \fn Reader::read_header( unsigned& nv , unsigned& nf )
   *
   * \brief Reads in  the number of vertices and  faces of the surface
   * described by the input file.
   *
   * \param nv A reference to the number of vertices of the surface.
   * \param nf A reference to the number of faces of the surface.
   */
  void
  Reader::read_header(
		      unsigned& nv,
		      unsigned& nf
                     )  
  {
    //
    // The first line should start with the string "OFF".
    //
    std::string stroff;

    assert( _is->get_string( stroff ) ) ;

    assert( stroff == "OFF" ) ;
   
    //
    // Read in the number of vertices of the mesh.
    //
    int number ;

    assert ( _is->get_integer( number ) ) ;

    nv = unsigned( number ) ;

    assert( nv >= 4 ) ;

    //
    // Read in the number of faces of the mesh.
    //

    assert( _is->get_integer( number ) ) ;

    nf = unsigned( number ) ;

    assert( nf >= 4 ) ;

    //
    // Read in the number of edges of the surface.
    //

    assert( _is->get_integer( number ) ) ;

    /**
     * There is no use for the integer "ne" here.
     */

    return ;
  }


  /**
   * \fn Reader::read_vertices( unsigned nv , double*& vset )
   *
   * \brief Reads  in the vertex coordinates of  the surface described
   * by the input file.
   *
   * \param nv A reference to the number of vertices of the surface.
   * \param vset A reference to an array of vertex coordinates.
   */
  void
  Reader::read_vertices(
			unsigned nv ,
			double*& vset
		       )
  {
    //
    // Allocate memory for storing the vertices.
    //
    vset = new double[ 3 * nv ] ;

    //
    // Read "nv" lines with the vertex coordinates
    // 
    for ( unsigned i = 0 ; i < nv ; i++ ) {
      //
      // Read in the first coordinate.
      //
      double x ;

      assert( _is->get_double( x ) ) ;

      //
      // Read in the second coordinate.
      //
      double y ;

      assert( _is->get_double( y ) ) ;

      //
      // Read in the third coordinate.
      //
      double z ;

      assert( _is->get_double( z ) ) ;

      const unsigned j = 3 * i ;

      vset[ j     ] = x ;
      vset[ j + 1 ] = y ;
      vset[ j + 2 ] = z ;
    }

    return ;
  }


  /**
   * \fn Reader::read_faces( unsigned nf , unsigned nv , unsigned*& fset )
   *
   * \brief  Reads  in the  face  vertex  identifiers  of the  surface
   * described by the input file.
   *
   * \param nf The number of faces of the surface.
   * \param nv The number of vertices of the surface.
   * \param fset A reference to an array of face vertex identifiers.
   */
  void
  Reader::read_faces(
		     unsigned nf,
		     unsigned nv,
		     unsigned*& fset
		    ) 
  {
    //
    // Allocate memory for storing the vertices.
    //
    fset = new unsigned[ 3 * nf ] ;

    //
    // Read "nf" lines of the form "3 v1 v2 v3", where "v1", "v2", and
    // "v3"  are the  vertex  indices  of the  face  defined by  these
    // vertices.
    //
    for ( unsigned i = 0 ; i < nf ; i++ ) {
      //
      // Read in the number of vertices of the i-th face.
      //
      int number ;

      assert( _is->get_integer( number ) ) ;

      assert( number == 3 ) ;

      //
      // Read in the identifier of the first face vertex.
      //

      assert( _is->get_integer( number ) ) ;

      unsigned i1 = unsigned( number ) ;      

      assert( i1 < nv ) ;

      //
      // Read in the identifier of the second face vertex.
      //

      assert( _is->get_integer( number ) ) ;

      unsigned i2 = unsigned( number ) ;      

      assert( i2 < nv ) ;

      //
      // Read in the identifier of the third face vertex.
      //

      assert( _is->get_integer( number ) ) ;

      unsigned i3 = unsigned( number ) ;      

      assert( i3 < nv ) ;

      //
      // Store face data.
      //
      const unsigned j = 3 * i ;

      fset[ j     ] = i1 ;
      fset[ j + 1 ] = i2 ;
      fset[ j + 2 ] = i3 ;
    }

    return ;
  }

}

/** @} */ //end of group class.
