/** 
 * \file reader.h
 *  
 * \brief  Definition of the  class Reader,  which represents  a file
 * reader for reading  in a triangle surface mesh  information from an
 * OFF file.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date July 2009
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef READER_H
#define READER_H

#include "lexer.h"

#include <fstream>
#include <string>


/** 
* \defgroup OFFLIBNameSpace Namespace offlib.
* @{
*/

/**
* \namespace offlib
*
* \brief   The   namespace   offlib   contains  the   definition   and
* implementation of  classes for  reading and writing  the topological
* and geometric information of triangle meshes from / to OFF files.
*/

namespace offlib {

  /**
   * \class Reader
   *
   * \brief  This class  represents a  file  reader for  reading in  a
   * triangle surface mesh information from an OFF file.
   */
  class Reader {
  public:
    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn Reader( const std::string& fn )
     *
     * \brief Creates an instance of this class.
     *
     * \param fn The name of an OFF input file.
     */
    Reader( const std::string& fn ) : _is( 0 )
    {
      _fname = fn + std::string( ".off" ) ;
    }


    /**
     * \fn ~Reader
     *
     * \brief Destroys an instance of this class.
     */
    ~Reader()
    {
      if ( _is != 0 ) delete _is ; 
    }


    /**
     * \fn read( unsigned& nv , double*& vset , unsigned& nf , unsigned*& fset ) 
     *
     * \brief Reads  in the  topological and geometric  information of
     * the surface mesh described in the input file.
     *
     * \param nv A reference to the number of vertices of the surface.
     * \param  vset   A  reference  to   an  array  with   the  vertex
     * coordinates.
     * \param nf A reference to the number of faces of the surface.
     * \param  fset A  reference  to  an array  with  the face  vertex
     * identifiers.
     */
    void read(
	      unsigned& nv ,
	      double*& vset ,
	      unsigned& nf ,
	      unsigned*& fset 
	     ) ; 


  private:

    // ---------------------------------------------------------------
    //
    // Private methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn read_header( unsigned& nv , unsigned& nf )
     *
     * \brief Reads in the number of vertices and faces of the surface
     * described by the input file.
     *
     * \param nv A reference to the number of vertices of the surface.
     * \param nf A reference to the number of faces of the surface.
     */
    void read_header( 
		     unsigned& nv , 
		     unsigned& nf 
		    ) ;


    /**
     * \fn read_vertices( unsigned nv , double*& vset )
     *
     * \brief Reads in the vertex coordinates of the surface described
     * by the input file.
     *
     * \param nv The number of vertices of the surface.
     * \param vset A reference to an array of vertex coordinates.
     */
    void read_vertices( 
		       unsigned nv , 
		       double*& vset 
		      ) ;


    /**
     * \fn read_faces( unsigned nf , unsigned nv , unsigned*& fset )
     *
     * \brief  Reads in  the face  vertex identifiers  of  the surface
     * described by the input file.
     *
     * \param nf The number of faces of the surface.
     * \param nv The number of vertices of the surface.
     * \param fset A reference to an array of face vertex identifiers.
     */
    void read_faces( 
		    unsigned nf , 
		    unsigned nv , 
		    unsigned*& fset 
		   ) ;


    // ---------------------------------------------------------------
    //
    // Private data members
    //
    // ---------------------------------------------------------------

    std::string _fname ;   ///< The input file name.

    Lexer* _is ;           ///< A lexical analyzer for the input file stream.

  } ;

} 

/** @} */ //end of group class.

#endif   // READER_H
