/** 
 * \file writer.cpp
 *  
 * \brief Implementation of the  class Writer, which represents a file
 * writer for  writing out a  triangle surface mesh information  to an
 * OFF file.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date July 2009
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include <iostream>
#include <iomanip>
#include <cassert>

#include "writer.h"


/** 
* \defgroup OFFLIBNameSpace Namespace offlib.
* @{
*/

/**
* \namespace offlib
*
* \brief   The   namespace   offlib   contains  the   definition   and
* implementation of  classes for  reading and writing  the topological
* and geometric information of triangle meshes from / to OFF files.
*/

namespace offlib {


  /**
   * \fn Writer::Writer( const std::string& fn )
   *
   * \brief Creates an instance of this class.
   *
   * \param fn The name of an OFF input file.
   */
  Writer::Writer( const std::string& fn )
  {
    assert( !fn.empty() ) ;
    
    _fname = fn + std::string( ".off" ) ;

    return ;
  }


  /**
   * \fn Writer::Writer( const Writer& w )
   *
   * \brief Creates an instance of this class from another instance.
   *
   * \param w Another instance of this class.
   */
  Writer::Writer( const Writer& w ) : _fname( w._fname )
  {}


  /**
   * \fn void Writer::write( unsigned nv , double* vset , unsigned nf , unsigned* fset )
   *
   * \brief Writes the vertex and face information of a surface mesh
   * to an OFF file.
   *
   * \param nv The number of vertices of the mesh.
   * \param vset The Cartesian coordinates of the mesh vertices.
   * \param nf The number of faces of the mesh.
   * \param fset The set of vertex identifiers of the mesh faces.
   */
  void
  Writer::write(
		unsigned nv ,
		double* vset ,
		unsigned nf ,
		unsigned* fset
	       )
  {
    _fs.open( _fname.c_str() , std::ios::out | std::ios::binary ) ;

    assert( _fs != 0 ) ;

    write_header( nv , nf ) ;

    write_vertices( nv , vset ) ;

    write_faces( nf , fset ) ;

    _fs.close() ;

    return ;
  }


  /**
   * \fn void Writer::write_header( unsigned nv , unsigned nf )
   *
   * \brief Writes the file header information.
   *
   * \param nv The number of vertices of the mesh.
   * \param nf The number of faces of the mesh.
   */
  void
  Writer::write_header(
		       unsigned nv ,
		       unsigned nf
		      )
  {
    //
    // Write first line
    //
    _fs << "OFF" << std::endl ;

    //
    // Write the number of vertices, faces, and edges
    //
    _fs << nv
	<< " "
	<< nf
	<< " 0"
	<< std::endl ;
  }


  /**
   * \fn void Writer::write_vertices( unsigned nv , double* vset )
   *
   * \brief Writes the mesh vertex information.
   *
   * \param nv The number of vertices of the mesh.
   * \param vset The Cartesian coordinates of the mesh vertices.
   */
  void
  Writer::write_vertices(
			 unsigned nv ,
			 double* vset
			) 
  {
    _fs << std::fixed << std::setprecision( 18 ) ;

    for ( unsigned i = 0 ; i < nv ; i++ ) {

      const unsigned j = 3 * i ;

      _fs << vset[ j ]
	   << " "
	   << vset[ j + 1 ] 
	   << " "
	   << vset[ j + 2 ]
	   << std::endl;
    }

    return ;
  }


  /**
   * \fn void Writer::write_faces( unsigned nf , unsigned* fset )
   *
   * \brief Writes the mesh face information.
   *
   * \param nf The number of faces of the mesh.
   * \param fset The set of vertex identifiers of the mesh faces.
   */
  void
  Writer::write_faces(
		      unsigned nf ,
		      unsigned* fset
		     ) 
  {
    for ( unsigned i = 0 ; i < nf ; i++ ) {

      const unsigned j = 3 * i;

      _fs << "3 "
	  << fset[ j ]
	  << " "
	  << fset[ j + 1 ]
	  << " "
	  << fset[ j + 2 ]
	  << std::endl ;
    }

    return ;
  }

}

/** @} */ //end of group class.
