/** 
 * \file test-adgl.cpp
 *  
 * \brief An application to perform  a simple and random test of class
 * Geodesics, which computes  discrete geodesics over triangle surface
 * meshes.
 *
 * \author
 * Mario Augusto de Souza Lizier \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * lizier at gmail (dot) com
 *
 * \author
 * Marcelo Ferreira Siqueira
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * mfsiqueira at gmail (dot) com
 *
 * \version 1.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include <iostream>                 // std::cout, std::endl
#include <string>                   // std::string
#include <cstdlib>                  // atoi, atof, srand, exit
#include <iomanip>                  // std:setprecision
#include <ctime>                    // time, clock, CLOCKS_PER_SEC, clock_t 
#include <fstream>                  // std::ifstream, std::ofstream

#include "surface.h"                // dcel::Surface

#include "geodesic_point.h"         // adg::Geodesic_point
#include "geodesic_point_vertex.h"  // adg::Geodesic_point_vertex
#include "geodesic_point_edge.h"    // adg::Geodesic_point_edge
#include "geodesic_point_face.h"    // adg::Geodesic_point_face
#include "curve_point.h"            // adg::Curve_point
#include "geodesics.h"              // adg::Geodesics

#include "adg_dcel_mesh.h"          // adgfromdcel::Adg_dcel_mesh
#include "adg_vertex_attributes.h"  // adgfromdcel::Adg_vertex_attributes

#include "reader.h"                 // offlib::Reader

using namespace std ;

/**
 * \typedef mySurface
 *
 * \brief Definition of a type name for the mesh data structure.
 */
typedef dcel::Surface< adgfromdcel::Adg_vertex_attributes > mySurface ;


/**
 * \typedef CM
 *
 * \brief Definition of a type name for the concrete mesh class.
 */
typedef adgfromdcel::Adg_dcel_mesh CM ;


/**
 * \typedef GP
 *
 * \brief Definition of a type name for the geodesic vertices.
 */
typedef adg::Geodesic_point< mySurface > GP ;


/**
 * \typedef GPV
 *
 * \brief Definition of a type name for the geodesic vertices that are
 * mesh vertices.
 */
typedef adg::Geodesic_point_vertex< mySurface > GPV ;


/**
 * \typedef GPE
 *
 * \brief Definition of a type name for the geodesic vertices that are
 * contained in the interior of mesh edges.
 */
typedef adg::Geodesic_point_edge< mySurface > GPE ;


/**
 * \typedef GPF
 *
 * \brief Definition of a type name for the geodesic vertices that are
 * contained in the interior of mesh faces.
 */
typedef adg::Geodesic_point_face< mySurface > GPF ;


/**
 * \typedef Geodesics
 *
 * \brief Definition of a type name for the geodesic class.
 */
typedef adg::Geodesics< mySurface > Geodesics ;


/**
 * \fn void write( CM* cm_mesh , std::string filename , std::vector< std::list< GP* > >& pts )
 *
 * \brief Writes the geodesic vertices to a VTK file.
 *
 * \param cm_mesh Pointer to the mesh over which the geodesics is defined.
 * \param filename The VTK file name.
 * \param pts  An array  with a  set of geodesics,  which of  which is
 * given as a list of its vertices.
 */
void write( CM* cm_mesh , std::string filename, std::vector< std::list< GP* > >& pts ) ;


/**
 * \fn void usage()
 *
 * \brief Provides  the user  with a brief  explanation on how  to use
 * this program.
 *
 */
void usage( void ) ;


/**
 * \fn int main( int argc , char* argv[] )
 *
 * \brief  Computes  a user  defined  set  of  geodesics over  a  user
 * provided surface mesh.
 *
 * \param argc The number of command line arguments.
 * \param argv The command line arguments.
 *
 * \return An integer number.
 */
int main( int argc, char* argv[] ) {
  using std::cout ;
  using std::endl ;

  if ( ( argc < 5 ) || ( argc > 11 ) ) {
    usage() ;
    return 0 ;
  }

  int mode = -1 ;
  unsigned v1 = -1 ;
  unsigned v2 = -1 ;
  double uu[ 2 ] = { 0 , 0 } ;
  double vv[ 2 ] = { 0 , 0 } ;
  double tt[ 2 ] = { 0 , 0 } ;
  char *source = 0 ;
  char *dest = 0 ;
  char *save = 0 ;
  char *load = 0 ;
  unsigned nrand = 0 ;

  /*
   * Parse the command-line arguments.
   */
  if ( 
      ( string( argv[ 1 ] ).compare( "-v" ) == 0 ) ||
      ( string( argv[ 1 ] ).compare( "-e" ) == 0 ) ||
      ( string( argv[ 1 ] ).compare( "-f" ) == 0 ) 
     ) {

    if ( ( argc < 7 ) || ( argc > 11 ) ) {
      usage() ;
      return 0 ;
    }

    if (
	( string( argv[ 1 ] ).compare( "-v" ) == 0 ) &&
	( string( argv[ 3 ] ).compare( "-v" ) != 0 ) &&
	( string( argv[ 3 ] ).compare( "-e" ) != 0 ) &&
	( string( argv[ 3 ] ).compare( "-f" ) != 0 ) 
       ) {
      usage() ;
      return 0 ;
    }

    if (
	( string( argv[ 1 ] ).compare( "-e" ) == 0 ) &&
	( string( argv[ 4 ] ).compare( "-v" ) != 0 ) &&
	( string( argv[ 4 ] ).compare( "-e" ) != 0 ) &&
	( string( argv[ 4 ] ).compare( "-f" ) != 0 ) 
       ) {
      usage() ;
      return 0 ;
    }

    if (
	( string( argv[ 1 ] ).compare( "-f" ) == 0 ) &&
	( string( argv[ 5 ] ).compare( "-v" ) != 0 ) &&
	( string( argv[ 5 ] ).compare( "-e" ) != 0 ) &&
	( string( argv[ 5 ] ).compare( "-f" ) != 0 ) 
       ) {
      usage() ;
      return 0 ;
    }

    v1 = ( unsigned ) atoi( argv[ 2 ] ) ;

    if ( string( argv[ 1 ] ).compare( "-v" ) == 0 ) {
      v2 = ( unsigned ) atoi( argv[ 4 ] ) ;
      if ( string( argv[ 3 ] ).compare( "-e" ) == 0 ) {
	tt[ 1 ] = ( double ) atof( argv[ 5 ] ) ;
	source = argv[ 6 ] ;
	dest = argv[ 7 ] ;
      }
      else if ( string( argv[ 3 ] ).compare( "-f" ) == 0 ) {
	uu[ 1 ] = ( double ) atof( argv[ 5 ] ) ;
	vv[ 1 ] = ( double ) atof( argv[ 6 ] ) ;
	source = argv[ 7 ] ;
	dest = argv[ 8 ] ;
      }
      else {
	source = argv[ 5 ] ;
	dest = argv[ 6 ] ;
      }
    }
    else if ( string( argv[ 1 ] ).compare( "-e" ) == 0 ) {
      v2 = ( unsigned ) atoi( argv[ 5 ] ) ;
      tt[ 0 ] = ( double ) atof( argv[ 3 ] ) ;
      if ( string( argv[ 4 ] ).compare( "-e" ) == 0 ) {
	tt[ 1 ] = ( double ) atof( argv[ 6 ] ) ;
	source = argv[ 7 ] ;
	dest = argv[ 8 ] ;
      }
      else if ( string( argv[ 4 ] ).compare( "-f" ) == 0 ) {
	uu[ 1 ] = ( double ) atof( argv[ 6 ] ) ;
	vv[ 1 ] = ( double ) atof( argv[ 7 ] ) ;
	source = argv[ 8 ] ;
	dest = argv[ 9 ] ;
      }
      else {
	source = argv[ 6 ] ;
	dest = argv[ 7 ] ;
      }
    }
    else {
      v2 = ( unsigned ) atoi( argv[ 6 ] ) ;
      uu[ 0 ] = ( double ) atof( argv[ 3 ] ) ;
      vv[ 0 ] = ( double ) atof( argv[ 4 ] ) ;
      if ( string( argv[ 5 ] ).compare( "-e" ) == 0 ) {
	tt[ 1 ] = ( double ) atof( argv[ 7 ] ) ;
	source = argv[ 8 ] ;
	dest = argv[ 9 ] ;
      }
      else if ( string( argv[ 5 ] ).compare( "-f" ) == 0 ) {
	uu[ 1 ] = ( double ) atof( argv[ 7 ] ) ;
	vv[ 1 ] = ( double ) atof( argv[ 8 ] ) ;
	source = argv[ 9 ] ;
	dest = argv[ 10 ] ;
      }
      else {
	source = argv[ 7 ] ;
	dest = argv[ 8 ] ;
      }
    }

    mode = 1 ;
  }
  else if ( string( argv[ 1 ] ).compare( "-r" ) == 0 ) {
    if ( ( argc != 5 ) && ( argc != 7 ) ) {
      usage() ;
      return 0 ;
    }

    nrand = atoi( argv[ 2 ] ) ;

    if ( argc == 7 ) {
      if ( string( argv[ 3 ] ).compare( "-s" ) == 0 ) {
        save = argv[ 4 ] ;
        source = argv[ 5 ] ;
        dest = argv[ 6 ] ;
      } 
      else {
        usage() ;
        return 0 ;
      }
    } 
    else {
      source = argv[ 3 ] ;
      dest = argv[ 4 ] ;
    }

    mode = 2 ;
  }
  else if ( string( argv[ 1 ] ).compare( "-l" ) == 0 ) {
    if ( argc != 5 ) {
      usage() ;
      return 0 ;
    }

    load = argv[ 2 ] ;
    source = argv[ 3 ] ;
    dest = argv[ 4 ] ;
    mode = 3 ;
  }

  if ( mode == -1 ) {
    usage() ;
    return 0 ;
  }

  unsigned nv, nf ;
  double* vset ;
  unsigned* fset ;

  clock_t start, end ;

  cout << "Reading input file..." ;
  cout.flush() ;

  std::string fn( source ) ;

  start = clock() ;
  offlib::Reader reader( fn ) ;
  reader.read( nv, vset, nf, fset ) ;
  end = clock() ;

  cout << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
       << " seconds." 
       << endl ;

  //
  // Creates the underlying mesh and stores it in an ADCEL.
  //

  cout << "Creating surface mesh... " ;
  cout.flush() ;

  start = clock() ;
  mySurface* mesh = new mySurface( nv, vset, nf, fset ) ;
  end = clock() ;

  cout << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
       << " seconds." 
       << endl ;

  //
  // Creates the mesh geodesic interface.
  //

  cout << "Creating the mesh data structure interface... " ;
  cout.flush() ;

  start = clock() ;
  CM* cm_mesh = new CM( mesh ) ;
  end = clock() ;

  cout << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
       << " seconds." 
       << endl ;

  cout << "Creating the initial and final geodesic vertices... " ;
  cout.flush() ;


  start = clock() ;
  std::vector< pair< GP* , GP* > > pairs ;

  srand( time( NULL ) ) ;

  if ( mode == 1 ) {
    unsigned index = 1 ;
    GP *vi ;
    if ( string( argv[ index ] ).compare( "-v" ) == 0 ) {
      if ( v1 > mesh->get_number_of_vertices() ) {
	cout << "Bad initial vertex index"
         << endl
         << endl ;
	return 0 ;
      }
      mySurface::VertexIterator iti = mesh->vertices_begin() ;
      for ( unsigned i = 0 ; i < v1 ; i++ ) {
	++iti ;
      }
      vi = new GPV( *iti ) ;
      index = 3 ; 
    }
    else if ( string( argv[ index ] ).compare( "-e" ) == 0 ) {
      if ( v1 > mesh->get_number_of_edges() ) {
	cout << "Bad initial edge index"
         << endl
         << endl ;
	return 0 ;
      }
      mySurface::EdgeIterator iti = mesh->edges_begin() ;
      for ( unsigned i = 0 ; i < v1 ; i++ ) {
	++iti ;
      }
      vi = new GPE( *iti , tt[ 0 ] ) ;
      index = 4 ; 
    }
    else {
     if ( v1 > mesh->get_number_of_faces() ) {
	cout << "Bad initial face index"
         << endl
         << endl ;
	return 0 ;
      }
      mySurface::FaceIterator iti = mesh->faces_begin() ;
      for ( unsigned i = 0 ; i < v1 ; i++ ) {
	++iti ;
      }
      vi = new GPF( *iti , uu[ 0 ] , vv[ 0 ] ) ;
      index = 5 ; 
    }

    GP *ve ;
    if ( string( argv[ index ] ).compare( "-v" ) == 0 ) {
      if ( v2 > mesh->get_number_of_vertices() ) {
	cout << "Bad initial vertex index"
         << endl
         << endl ;
	return 0 ;
      }
      mySurface::VertexIterator itf = mesh->vertices_begin() ;
      for ( unsigned i = 0 ; i < v2 ; i++ ) {
	++itf ;
      }
      ve = new GPV( *itf ) ;
    }
    else if ( string( argv[ index ] ).compare( "-e" ) == 0 ) {
      if ( v2 > mesh->get_number_of_edges() ) {
	cout << "Bad initial edge index"
         << endl
         << endl ;
	return 0 ;
      }
      mySurface::EdgeIterator itf = mesh->edges_begin() ;
      for ( unsigned i = 0 ; i < v2 ; i++ ) {
	++itf ;
      }
      ve = new GPE( *itf , tt[ 1 ] ) ;
    }
    else {
     if ( v2 > mesh->get_number_of_faces() ) {
	cout << "Bad initial face index"
         << endl
         << endl ;
	return 0 ;
      }
      mySurface::FaceIterator itf = mesh->faces_begin() ;
      for ( unsigned i = 0 ; i < v2 ; i++ ) {
	++itf ;
      }
      ve = new GPF( *itf , uu[ 1 ] , vv[ 1 ] ) ;
    }

    pairs.push_back( pair< GP* , GP* > ( vi , ve ) ) ;
  }
  else if ( mode == 2 ) {
    if ( nrand <= 0 ) {
      cout << "The number of geodesics must be larger than 0" 
	   << endl
	   << endl ;
      
      return 0 ;
    }

    std::ofstream pf ;

    if ( save != 0 ) {
      pf.open( string( string( save ) + string( ".txt" ) ).c_str() ) ;
      if ( pf.is_open() ) {
	pf << nrand << endl ;
      } 
    }

    for ( unsigned p = 0 ; p < nrand ; p++ ) {
      GP* gp[ 2 ] ;
      for ( int i = 0 ; i < 2 ; i++ ) {
	unsigned nv = rand() % mesh->get_number_of_vertices() ;
	if ( nv > mesh->get_number_of_vertices() ) {
	  cout << "Bad initial vertex index"
	       << endl
	       << endl ;
	  return 0 ;
	}
	
	mySurface::VertexIterator itv = mesh->vertices_begin() ;
	for ( unsigned j = 0 ; j < nv ; j++ ) {
	  ++itv ;
	}

	gp[ i ] = new GPV( *itv ) ;

	if ( save != 0 ) {
	  if ( i == 0 ) {
	    pf << nv << '\t' ;
	  }
	  else {
	    pf << nv << endl ;
	  }
	}
      }

      pairs.push_back( pair< GP* , GP* > ( gp[ 0 ] , gp[ 1 ] ) ) ;
    }

    if ( save != 0 ) {
      if ( pf.is_open() ) {
	pf.close() ;
      } 
    }
  }
  else {
    std::ifstream pf( string ( string( load ) + string( ".txt" ) ).c_str() ) ;

    if ( pf.is_open() ) {
      unsigned npairs ;
      pf >> npairs ;

      if ( npairs <= 0 ) {
        cout << "The number of pairs must be greater than zero." 
	     << endl ;
        return 0  ;
      }

      for ( unsigned p = 0 ; p < npairs ; p++ ) {
	GP* gp[ 2 ] ;
	for ( int i = 0 ; i < 2 ; i++ ) {
	  unsigned nv ;
	  pf >> nv ;

	  if ( nv > mesh->get_number_of_vertices() ) {
	    cout << "Bad initial vertex index"
		 << endl
		 << endl ;
	    return 0 ;
	  }

	  mySurface::VertexIterator itv = mesh->vertices_begin() ;
	  for ( unsigned j = 0 ; j < nv ; j++ ) {
	    ++itv ;
	  }

	  gp[ i ] = new GPV( *itv ) ;
	}

	pairs.push_back( pair< GP* , GP* > ( gp[ 0 ] , gp[ 1 ] ) ) ;
      }
    }
    
    pf.close() ;
  }
 
  end = clock() ;

  cout << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
       << " seconds." 
       << endl ;

  //
  // Create an instance of a geodesic.
  //  
  Geodesics geo( cm_mesh, 1e-12, 1e6 ) ;

  std::vector< std::list< GP* > > paths_first ;
  std::vector< std::list< GP* > > paths ;

  for ( unsigned p = 0; p < pairs.size(); p++ ) {
    cout << "Searching vertices... " ;
    cout.flush() ;

    start = clock() ;

    GP* vi = pairs[ p ].first ;
    GP* ve = pairs[ p ].second ;

    end = clock() ;

    cout << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
	 << " seconds."
	 << endl ;

    cout << "Calculating geodesic " 
	 << p + 1
	 << " ... " ;

    cout.flush() ;

    start = clock() ;
    std::list< GP* > pts_first ;
    std::list< GP* > pts ;
    geo.compute_first_approximation( vi , ve , pts_first ) ;
    geo.shorten_geodesic( pts_first , pts ) ;
    end = clock() ;

    cout << "Iterations: " 
	 << geo.get_total_number_of_iterations()
         << "... " ;

    cout << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
	 << " seconds."
	 << endl ;

    std::list< GP* > pts_first_temp = pts_first ;
    paths_first.push_back( pts_first_temp ) ;
    pts_first.clear() ;

    std::list< GP* > pts_temp = pts ;
    paths.push_back( pts_temp ) ;
    pts.clear() ;
  }

  //
  // Writing to the output file
  //

  cout << "Writing output files... " ;
  cout.flush() ;

  start = clock() ;
  write( 
	cm_mesh , 
	std::string( dest ) + std::string( "-first.vtk" ) ,
	paths_first
       ) ;

  write( 
	cm_mesh , 
	std::string( dest ) + std::string( ".vtk" ) ,
	paths 
       ) ;
  end = clock() ;

  cout << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
       << " seconds." 
       << endl ;

  //
  // Release memory.
  // 

  cout << "Releasing memory..." << endl ;
  
  if ( vset != 0 ) {
    delete[] vset ;
  }

  if ( fset != 0 ) {
    delete[] fset ;
  }

  if ( mesh != 0 ) {
    delete mesh ;
  }

  if ( cm_mesh != 0 ) {
    delete cm_mesh ;
  }

  cout << "Finished." << endl ;
  
  return 0 ;
}


/**
 * \fn void write( CM* cm_mesh , std::string filename , std::vector< std::list< GP* > >& pts )
 *
 * \brief Writes the geodesic vertices to a VTK file.
 *
 * \param cm_mesh Pointer to the mesh over which the geodesics is defined.
 * \param filename The VTK file name.
 * \param pts  An array  with a  set of geodesics,  which of  which is
 * given as a list of its vertices.
 */
void write(
	   CM* cm_mesh ,
	   std::string filename ,
	   std::vector< std::list< GP* > >& pts 
	  ) 
{
  std::ofstream pf( filename.c_str() ) ;

  if ( pf.is_open() ) {

    pf << "# vtk DataFile Version 1.0" << std::endl ;
    pf << "Mesh from mf" << std::endl ;
    pf << "ASCII" << std::endl ;
    pf << "DATASET POLYDATA" << std::endl ;

    int npoints = 0 ;
    for ( unsigned p = 0 ; p < pts.size() ; p++ )
      npoints += pts[ p ].size() ;

    pf << "POINTS " << npoints << " float" << std::endl ;

    for ( unsigned p = 0; p < pts.size() ; p++ ) {
      for ( std::list< GP* >::iterator itv = pts[ p ].begin() ;
	    itv != pts[ p ].end() ; ++itv ) {
        double x, y, z ;

        (*itv)->get_coords( cm_mesh , x , y , z ) ;

        pf << std::setprecision( 15 ) << x << " " ;
        pf << std::setprecision( 15 ) << y << " " ;
        pf << std::setprecision( 15 ) << z << std::endl ;
      }
    }

    pf << std::endl 
       << "LINES " 
       << pts.size() 
       << " " 
       << npoints + pts.size()
       << std::endl ;

    int count = 0 ;

    for ( unsigned p = 0 ; p < pts.size() ; p++ ) {
      pf << pts[ p ].size() ;

      for ( std::list< GP* >::iterator itv = pts[p].begin() ;
	    itv != pts[ p ].end() ; ++itv ) {
        pf << " " 
	   << count ;
        count++ ;
      }

      pf << std::endl ;
    }

    pf << std::endl ;

    pf << "CELL_DATA " << pts.size() << std::endl ;
    pf << "SCALARS colors int 1" << std::endl ;
    pf << "LOOKUP_TABLE default" << std::endl ;

    int color = 0 ;
    for ( unsigned p = 0 ; p < pts.size() ; p++ ) {
      pf << color 
	 << std::endl ;
      color++ ;
    }

    pf << std::endl ;

    pf.close() ;
  }

  return ;
}


/**
 * \fn void usage()
 *
 * \brief Provides  the user  with a brief  explanation on how  to use
 * this program.
 *
 */
void usage( void ) {
  cout << endl ;
  cout << "Modes:" 
       << endl ;
  cout << endl ;
  cout << "Calculate a single geodesic:" 
       << endl ;
  cout << "\t Usage: test-adgl -< v | e | f > <arg1> [ <arg2> ] [ <arg3> ] -< v | e | f > <arg4> [ <arg5> ] [ <arg6> ] <source> <dest>" 
       << endl ;
  cout << "\t\t arg1     the number of the first vertex, edge, or face" 
       << endl ;
  cout << "\t\t arg2     the interpolation coefficient if the first vertex is an edge or"
       << "\t\t          the first barycentric coordinate if the first vertex is a face" 
       << endl ;
  cout << "\t\t arg3     the second barycentric coordinate if the first vertex is a face" 
       << endl ;
 cout <<  "\t\t arg4     the number of the second vertex, edge, or face" 
       << endl ;
  cout << "\t\t arg5     the interpolation coefficient if the second vertex is an edge or"
       << "\t\t          the first barycentric coordinate if the second vertex is a face" 
       << endl ;
  cout << "\t\t arg6     the second barycentric coordinate if the second vertex is a face" 
       << endl ;
  cout << "\t\t source   the filename of an OFF file without extension" 
       << endl ;
  cout << "\t\t dest     the filename of an output VTK file without extension"
       << endl ;
  cout << endl << endl ;
  cout << endl << endl ;
  cout << "Calculate a given number of random geodesics from vertex to vertex:" 
       << endl ;
  cout << "\t Usage: test-adgl -r <arg1> [ -s <file> ] <source> <dest> " 
       << endl ;
  cout << "\t\t arg1     the number of the random pairs of initial and final points" 
       << endl ;
  cout << "\t\t -s       ask to save a txt file with all geodesic initial and final points"
       << endl ;
  cout << "\t\t file     the filename of a TXT file without extension" 
       << endl ;
  cout << "\t\t source   the filename of an OFF file without extension" 
       << endl ;
  cout << "\t\t dest     the filename of an output VTK file without extension"
       << endl ;

  cout << endl << endl ;

  cout << "Calculate all geodesics specified in a  file:" 
       << endl ;
  cout << "\t Usage: test-adgl -l <file> <source> <dest> " 
       << endl ;
  cout << "\t\t arg1     the number of the random pairs of initial and final points" 
       << endl ;
  cout << "\t\t file     the filename of a TXT file without extension" 
       << endl ;
  cout << "\t\t source   the filename of an OFF file without extension" 
       << endl ;
  cout << "\t\t dest     the filename of an output VTK file without extension"
       << endl ;
  cout << endl ;
  cout << endl ;

  return ;
}
